@@ -2,7 +2,6 @@ require 'reline/kill_ring'
 require 'reline/unicode'
 
 require 'tempfile'
-require 'pathname'
 
 class Reline::LineEditor
   # TODO: undo
@@ -51,18 +50,20 @@ class Reline::LineEditor
   CompletionJourneyData = Struct.new('CompletionJourneyData', :preposing, :postposing, :list, :pointer)
   MenuInfo = Struct.new('MenuInfo', :target, :list)
 
-  CSI_REGEXP = /\e\[[\d;]*[ABCDEFGHJKSTfminsuhl]/
-  OSC_REGEXP = /\e\]\d+(?:;[^;]+)*\a/
-  NON_PRINTING_START = "\1"
-  NON_PRINTING_END = "\2"
-  WIDTH_SCANNER = /\G(?:#{NON_PRINTING_START}|#{NON_PRINTING_END}|#{CSI_REGEXP}|#{OSC_REGEXP}|\X)/
-
   def initialize(config, encoding)
     @config = config
     @completion_append_character = ''
     reset_variables(encoding: encoding)
   end
 
+  def simplified_rendering?
+    if finished?
+      false
+    else
+      not @rerender_all and not finished? and Reline::IOGate.in_pasting?
+    end
+  end
+
   private def check_multiline_prompt(buffer, prompt)
     if @vi_arg
       prompt = "(arg: #{@vi_arg}) "
@@ -73,14 +74,39 @@ class Reline::LineEditor
     else
       prompt = @prompt
     end
+    return [prompt, calculate_width(prompt, true), [prompt] * buffer.size] if simplified_rendering?
     if @prompt_proc
       prompt_list = @prompt_proc.(buffer)
       prompt_list.map!{ prompt } if @vi_arg or @searching_prompt
+      if @config.show_mode_in_prompt
+        if @config.editing_mode_is?(:vi_command)
+          mode_icon = @config.vi_cmd_mode_icon
+        elsif @config.editing_mode_is?(:vi_insert)
+          mode_icon = @config.vi_ins_mode_icon
+        elsif @config.editing_mode_is?(:emacs)
+          mode_icon = @config.emacs_mode_string
+        else
+          mode_icon = '?'
+        end
+        prompt_list.map!{ |pr| mode_icon + pr }
+      end
       prompt = prompt_list[@line_index]
       prompt_width = calculate_width(prompt, true)
       [prompt, prompt_width, prompt_list]
     else
       prompt_width = calculate_width(prompt, true)
+      if @config.show_mode_in_prompt
+        if @config.editing_mode_is?(:vi_command)
+          mode_icon = @config.vi_cmd_mode_icon
+        elsif @config.editing_mode_is?(:vi_insert)
+          mode_icon = @config.vi_ins_mode_icon
+        elsif @config.editing_mode_is?(:emacs)
+          mode_icon = @config.emacs_mode_string
+        else
+          mode_icon = '?'
+        end
+        prompt = mode_icon + prompt
+      end
       [prompt, prompt_width, nil]
     end
   end
@@ -116,7 +142,7 @@ class Reline::LineEditor
           if @line_index.zero?
             0
           else
-            calculate_height_by_lines(@buffer_of_lines[0..(@line_index - 1)], prompt_list)
+            calculate_height_by_lines(@buffer_of_lines[0..(@line_index - 1)], prompt_list || prompt)
           end
         if @prompt_proc
           prompt = prompt_list[@line_index]
@@ -190,10 +216,10 @@ class Reline::LineEditor
     @is_multiline = false
   end
 
-  private def calculate_height_by_lines(lines, prompt_list)
+  private def calculate_height_by_lines(lines, prompt)
     result = 0
+    prompt_list = prompt.is_a?(Array) ? prompt : nil
     lines.each_with_index { |line, i|
-      prompt = ''
       prompt = prompt_list[i] if prompt_list and prompt_list[i]
       result += calculate_height_by_width(calculate_width(prompt, true) + calculate_width(line))
     }
@@ -211,40 +237,8 @@ class Reline::LineEditor
     width.div(@screen_size.last) + 1
   end
 
-  private def split_by_width(prompt, str, max_width)
-    lines = [String.new(encoding: @encoding)]
-    height = 1
-    width = 0
-    rest = "#{prompt}#{str}".encode(Encoding::UTF_8)
-    in_zero_width = false
-    rest.scan(WIDTH_SCANNER) do |gc|
-      case gc
-      when NON_PRINTING_START
-        in_zero_width = true
-      when NON_PRINTING_END
-        in_zero_width = false
-      when CSI_REGEXP, OSC_REGEXP
-        lines.last << gc
-      else
-        unless in_zero_width
-          mbchar_width = Reline::Unicode.get_mbchar_width(gc)
-          if (width += mbchar_width) > max_width
-            width = mbchar_width
-            lines << nil
-            lines << String.new(encoding: @encoding)
-            height += 1
-          end
-        end
-        lines.last << gc
-      end
-    end
-    # The cursor moves to next line in first
-    if width == max_width
-      lines << nil
-      lines << String.new(encoding: @encoding)
-      height += 1
-    end
-    [lines, height]
+  private def split_by_width(str, max_width)
+    Reline::Unicode.split_by_width(str, max_width, @encoding)
   end
 
   private def scroll_down(val)
@@ -312,6 +306,11 @@ class Reline::LineEditor
     @byte_pointer = new_byte_pointer
   end
 
+  def rerender_all
+    @rerender_all = true
+    rerender
+  end
+
   def rerender
     return if @line.nil?
     if @menu_info
@@ -358,7 +357,7 @@ class Reline::LineEditor
         new_lines = whole_lines
       end
       prompt, prompt_width, prompt_list = check_multiline_prompt(new_lines, prompt)
-      all_height = calculate_height_by_lines(new_lines, prompt_list)
+      all_height = calculate_height_by_lines(new_lines, prompt_list || prompt)
       diff = all_height - @highest_in_all
       move_cursor_down(@highest_in_all - @first_line_started_from - @started_from - 1)
       if diff > 0
@@ -398,7 +397,7 @@ class Reline::LineEditor
         if @line_index.zero?
           0
         else
-          calculate_height_by_lines(@buffer_of_lines[0..(@line_index - 1)], prompt_list)
+          calculate_height_by_lines(@buffer_of_lines[0..(@line_index - 1)], prompt_list || prompt)
         end
       if @prompt_proc
         prompt = prompt_list[@line_index]
@@ -457,7 +456,7 @@ class Reline::LineEditor
         if @line_index.zero?
           0
         else
-          calculate_height_by_lines(new_buffer[0..(@line_index - 1)], prompt_list)
+          calculate_height_by_lines(new_buffer[0..(@line_index - 1)], prompt_list || prompt)
         end
       @started_from = calculate_height_by_width(prompt_width + @cursor) - 1
       move_cursor_down(@first_line_started_from + @started_from)
@@ -488,7 +487,7 @@ class Reline::LineEditor
   end
 
   private def render_partial(prompt, prompt_width, line_to_render, with_control = true)
-    visual_lines, height = split_by_width(prompt, line_to_render.nil? ? '' : line_to_render, @screen_size.last)
+    visual_lines, height = split_by_width(line_to_render.nil? ? prompt : prompt + line_to_render, @screen_size.last)
     if with_control
       if height > @highest_in_this
         diff = height - @highest_in_this
@@ -507,8 +506,18 @@ class Reline::LineEditor
     Reline::IOGate.move_cursor_column(0)
     visual_lines.each_with_index do |line, index|
       if line.nil?
-        if Reline::IOGate.win? and calculate_width(visual_lines[index - 1], true) == Reline::IOGate.get_screen_size.last
-          # A newline is automatically inserted if a character is rendered at eol on command prompt.
+        if calculate_width(visual_lines[index - 1], true) == Reline::IOGate.get_screen_size.last
+          # reaches the end of line
+          if Reline::IOGate.win?
+            # A newline is automatically inserted if a character is rendered at
+            # eol on command prompt.
+          else
+            # When the cursor is at the end of the line and erases characters
+            # after the cursor, some terminals delete the character at the
+            # cursor position.
+            move_cursor_down(1)
+            Reline::IOGate.move_cursor_column(0)
+          end
         else
           Reline::IOGate.erase_after_cursor
           move_cursor_down(1)
@@ -528,22 +537,25 @@ class Reline::LineEditor
       end
     end
     Reline::IOGate.erase_after_cursor
+    Reline::IOGate.move_cursor_column(0)
     if with_control
-      move_cursor_up(height - 1)
+      # Just after rendring, so the cursor is on the last line.
       if finished?
-        move_cursor_down(@started_from)
+        Reline::IOGate.move_cursor_column(0)
+      else
+        # Moves up from bottom of lines to the cursor position.
+        move_cursor_up(height - 1 - @started_from)
+        Reline::IOGate.move_cursor_column((prompt_width + @cursor) % @screen_size.last)
       end
-      move_cursor_down(@started_from)
-      Reline::IOGate.move_cursor_column((prompt_width + @cursor) % @screen_size.last)
     end
     height
   end
 
   private def modify_lines(before)
-    return before if before.nil? || before.empty?
+    return before if before.nil? || before.empty? || simplified_rendering?
 
     if after = @output_modifier_proc&.call("#{before.join("\n")}\n", complete: finished?)
-      after.lines("\n", chomp: true)
+      after.lines("\n").map { |l| l.chomp('') }
     else
       before
     end
@@ -568,7 +580,7 @@ class Reline::LineEditor
       else
         i&.start_with?(target)
       end
-    }
+    }.uniq
     if is_menu
       menu(target, list)
       return nil
@@ -839,7 +851,7 @@ class Reline::LineEditor
     unless completion_occurs
       @completion_state = CompletionState::NORMAL
     end
-    if @is_multiline and @auto_indent_proc
+    if @is_multiline and @auto_indent_proc and not simplified_rendering?
       process_auto_indent
     end
   end
@@ -1041,6 +1053,7 @@ class Reline::LineEditor
 
   def finish
     @finished = true
+    @rerender_all = true
     @config.reset
   end
 
@@ -1058,29 +1071,7 @@ class Reline::LineEditor
   end
 
   private def calculate_width(str, allow_escape_code = false)
-    if allow_escape_code
-      width = 0
-      rest = str.encode(Encoding::UTF_8)
-      in_zero_width = false
-      rest.scan(WIDTH_SCANNER) do |gc|
-        case gc
-        when NON_PRINTING_START
-          in_zero_width = true
-        when NON_PRINTING_END
-          in_zero_width = false
-        when CSI_REGEXP, OSC_REGEXP
-        else
-          unless in_zero_width
-            width += Reline::Unicode.get_mbchar_width(gc)
-          end
-        end
-      end
-      width
-    else
-      str.encode(Encoding::UTF_8).grapheme_clusters.inject(0) { |w, gc|
-        w + Reline::Unicode.get_mbchar_width(gc)
-      }
-    end
+    Reline::Unicode.calculate_width(str, allow_escape_code)
   end
 
   private def key_delete(key)
@@ -1258,7 +1249,7 @@ class Reline::LineEditor
             if search_word.empty? and Reline.last_incremental_search
               search_word = Reline.last_incremental_search
             end
-            if @history_pointer # TODO
+            if @history_pointer
               case prev_search_key
               when "\C-r".ord
                 history_pointer_base = 0
@@ -2121,7 +2112,7 @@ class Reline::LineEditor
       fp.path
     }
     system("#{ENV['EDITOR']} #{path}")
-    @line = Pathname.new(path).read
+    @line = File.read(path)
     finish
   end
 
@@ -2321,7 +2312,6 @@ class Reline::LineEditor
     new_pointer = [@byte_pointer, @line_index]
     @previous_line_index = @line_index
     @byte_pointer, @line_index = @mark_pointer
-    @byte_pointer, @line_index = @mark_pointer
     @cursor = calculate_width(@line.byteslice(0, @byte_pointer))
     @cursor_max = calculate_width(@line)
     @mark_pointer = new_pointer
